# -*- encoding: utf-8 -*-
def main():
    code = bytearray([
        SPLIT, INSL, SEP, EXP, END
    ])
    vm = VM()
    result = spawn(vm, code, 0, (None, None))
    run(vm)
    print wait_for(vm, result)

(SPLIT,                 # a ⊗ a → a ⊗ a
 ASSL,                  # a ⊗ (b ⊗ c) → (a ⊗ b) ⊗ c
 ASSR,                  # (a ⊗ b) ⊗ c → a ⊗ (b ⊗ c)
 INSL,                  # a     → 1 ⊗ a
 DELL,                  # 1 ⊗ a → a
 EXCH,                  # a ⊗ b → b ⊗ a
 APP,                   # (a ⇒ b) ⊗ a → b
 SEL,                   # a & ... → a
 IND,                   # a → a ⊕ ...
 INST,                  # !a → a
 DUP,                   # !a → !a ⊗ !a
 DROP,                  # !a → 1
 ALT,                   # a → f a ⊕ g a
 OPT,                   # a → f a & g a
 CUR,                   # a → (b ⇒ f (b ⊗ a))
 EXP,                   # a → !(f a)
 SEP, END) = range(1, 19)

# data = unit | cons data data | comp data code | susp
def cons(a, b):
    return (a, b)

# Exponents are bit different from other computation.
def comp(sp, ref):
    if not isinstance(sp, list) or is_comp(sp, EXP):
        sp = [sp]
    sp.append(ref)
    return sp

def is_cons(sp):
    return isinstance(sp, tuple)

def is_comp(sp, sx):
    if isinstance(sp, list):
        code, pc = sp[-1]
        return code[pc] == sx
    else:
        return False

def is_susp(sp):
    return isinstance(sp, Susp)

def uncomp(sp):
    if len(sp) == 2:
        return sp[0], sp[1]
    else:
        ref = sp.pop()
        return sp, ref

class Susp(object):
    def __init__(self):
        self.state = EMPTY
        self.value = None

    def __repr__(self):
        if self.state == EMPTY:
            return "{?}"
        else:
            return "{" + repr(self.value) + "}"

class VM(object):
    def __init__(self):
        self.coroutines = []

def spawn(vm, code, pc, sp):
    result = Susp()
    vm.coroutines.append((result, code, pc, sp))
    return result

def run(vm):
    running = True
    while running and len(vm.coroutines) > 0:
        susp, code, pc, sp = vm.coroutines.pop(0)
        top = len(code)
        while pc < top and code[pc] != END and code[pc] != SEP:
            if code[pc] == SPLIT and is_cons(sp):
                sep_pc = forward_after(code, pc+1, SEP)
                end_pc = forward_after(code, sep_pc, END)
                s0 = Susp()
                s1 = Susp()
                vm.coroutines.append((s0, code, pc+1, sp[0]))
                vm.coroutines.append((s1, code, sep_pc, sp[1]))
                vm.coroutines.append((susp, code, end_pc, (s0,s1)))
                susp, code, pc, sp = vm.coroutines.pop(0)
            elif code[pc] == ASSL and is_cons(sp) and is_cons(sp[1]):
                (a, (b,c)) = sp
                sp = ((a,b), c)
            elif code[pc] == ASSR and is_cons(sp) and is_cons(sp[0]):
                ((a,b), c) = sp
                sp = (a, (b,c))
            elif code[pc] == INSL:
                sp = (None, sp)
                pc += 1
            elif code[pc] == DELL and is_cons(sp) and sp[0] is None:
                sp = sp[1]
            elif code[pc] == EXCH and is_cons(sp):
                (a, b) = sp
                sp = (b, a)
            elif code[pc] == APP and is_cons(sp) and is_comp(sp[0], CUR):
                sp, a = sp
                sp, (refcode, refpc) = uncomp(sp)
                s0 = Susp()
                vm.coroutines.append((s0, refcode, refpc+1, (sp, a)))
                end_pc = forward_after(code, sep_pc, END)
                # Just add the handle into suspension?
                vm.coroutines.append((susp, code, end_pc, s0))
                susp, code, pc, sp = vm.coroutines.pop(0)
            elif code[pc] == SEL and is_comp(sp, OPT):
                index, pc = getnum(code, pc+1)
                sp, (refcode, refpc) = uncomp(sp)
                refpc += 1
                for _ in range(index):
                    refpc = forward_after(refcode, refpc, SEP)
                s0 = Susp()
                vm.coroutines.append((s0, refcode, refpc, sp))
                # Just add the handle into suspension?
                vm.coroutines.append((susp, code, pc, s0))
                susp, code, pc, sp = vm.coroutines.pop(0)
            elif code[pc] == IND:
                sp = comp(sp, (code, pc))
                _, pc = getnum(code, pc+1)
            elif code[pc] == ALT and is_comp(sp, IND):
                sp, (refcode, refpc) = uncomp(sp)
                index, refpc = getnum(refcode, refpc+1)
                pc += 1
                for _ in range(index):
                    pc = forward_after(code, pc, SEP)
                end_pc = forward_after(code, pc, END)
                s0 = Susp()
                vm.coroutines.append((s0, code, pc, sp))
                # Some sort of stack would be useful?
                vm.coroutines.append((susp, code, end_pc, s0))
                susp, code, pc, sp = vm.coroutines.pop(0)
            elif code[pc] == INST and is_comp(sp, EXP):
                sp, (refcode, refpc) = uncomp(sp)
                s0 = Susp()
                vm.coroutines.append((s0, refcode, refpc+1, sp))
                vm.coroutines.append((susp, code, pc+1, s0))
                susp, code, pc, sp = vm.coroutines.pop(0)
            elif code[pc] == DUP and is_comp(sp, EXP):
                sp = (sp, sp)
            elif code[pc] == DROP and is_comp(sp, EXP):
                sp = None
            elif code[pc] == EXP or code[pc] == CUR or code[pc] == OPT:
                sp = comp(sp, (code, pc))
                pc = forward_after(code, pc+1, END)
            elif is_cons(sp) and is_susp(sp[0]):
                tg = sp[0]
                if tg.state == FILLED:
                    sp = (tg.value, sp[1])
                else:
                    assert tg.state == EMPTY, "user error"
                    tg.state = DELAYED
                    tg.value = (susp, code, pc, sp)
                    susp, code, pc, sp = vm.coroutines.pop(0)
            elif is_cons(sp) and is_susp(sp[1]):
                tg = sp[1]
                if tg.state == FILLED:
                    sp = (sp[0], tg.value)
                else:
                    assert tg.state == EMPTY, "user error"
                    tg.state = DELAYED
                    tg.value = (susp, code, pc, sp)
                    susp, code, pc, sp = vm.coroutines.pop(0)
            elif is_susp(sp):
                tg = sp
                if tg.state == FILLED:
                    sp = tg.value
                else:
                    assert tg.state == EMPTY, "user error"
                    tg.state = DELAYED
                    tg.value = (susp, code, pc, sp)
                    susp, code, pc, sp = vm.coroutines.pop(0)
            else:
                assert "type error"
        if susp.state == EMPTY:
            susp.state = FILLED
            susp.value = sp
        else:
            assert susp.state == DELAYED, "user error"
            if susp.value is None:
                running = False
                susp.state = FILLED
                susp.value = sp
            else:
                vm.coroutines.append(susp.value)
                susp.state = FILLED
                susp.value = sp

EMPTY = 0
FILLED = 1
DELAYED = 2

def wait_for(vm, susp):
    if susp.state == EMPTY:
        susp.state = DELAYED
        assert susp.value == None
        while susp.state == DELAYED:
            run(vm)
    else:
        assert susp.state == FILLED, "user error"
    return susp.value

def getnum(code, pc):
    octet = code[pc]
    pc += 1
    sign = octet & 64
    num = (octet & 63) - (octet & sign)
    while octet >= 128:
        octet = code[pc]
        pc += 1
        num = (num << 7) | (octet & 127)
    return num, pc

def forward_after(code, pc, octet):
    depth = 0
    while depth > 0 or code[pc] != octet:
        if octet == END:
            depth -= 1
        if octet in (SPLIT, ALT, OPT, CUR, EXP):
            depth += 1
        pc += 1
    assert code[pc] == octet, "type error"
    pc += 1
    return pc

if __name__=="__main__":
    main()
